#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef LINUX
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#else
#include <windows.h>
#include <winsock2.h>
#endif

#include "ethernet.h"
#include "commands.h"
#include "errmsgs.h"
#include "md5.h"
#include "crc32.h"

// for the password authentication a challenge response method is used
#define SIZEOFCHALLENGE	16
#define PORT 54131                      // udp and tcp port of the devices

#define SIZE_MAGICNUMBER	4
char magicnumber[SIZE_MAGICNUMBER] = "dplc";
#define SWAPPED_EEPROM_SIGNATURE	0x636c7064

#define BUFF_SIZE 			256     		 // input buffer

#define CHIP_NAME_LEN      8
#define FW_VERS_LEN        6
#define SERIAL_NUM_LEN     24
#define UDP_TIMEOUT			500 			// time in ms that the programm is waiting for responses of the devices

#ifdef LINUX
#define SOCKET int
#define SOCKADDR_IN struct sockaddr_in
#define SOCKADDR struct sockaddr
#define SOCKLEN socklen_t
#define INVALID_SOCKET  (SOCKET)(~0)
#define SOCKET_ERROR -1
#else
#define SOCKLEN int
#endif


SOCKET udpsocket, tcpsocket;
struct sockaddr_in sendAddr;
int sendAddrLen;

// via udp the device respons with this structure
// it's important that the structure must be packed!

#ifdef LINUX
	#pragma pack(1)
#else
	#pragma option -a1
#endif

typedef struct
{
	dword magic_no;					// 0x626c7064
	byte struct_vers;					// 0x01
	byte struct_len;					// length of structure	
	byte hardware_revision;			// in BCD
	char chip_name[CHIP_NAME_LEN];// name of chip (e.g. "DPC3080")
	char fw_version[FW_VERS_LEN];	// firmware version (e.g. "2.04")
	char serno[SERIAL_NUM_LEN];	// serial number
	byte pw_mode;						// password mode (0x00 - 0x03)
} tcp_info_1;

typedef struct
{
   dword magic_no;					// 0x626c7064
	byte struct_vers;					// 0x02
	byte struct_len;					// length of structure	
	byte hardware_revision;			// in BCD
	char chip_name[CHIP_NAME_LEN];// name of chip (e.g. "DPC3080")
	char fw_version[FW_VERS_LEN];	// firmware version (e.g. "2.04")
	char serno[SERIAL_NUM_LEN];	// serial number
	byte pw_mode;						// password mode (0x00 - 0x03)
	dword restricted_ip;				// ip to which the iLCD is restricted
											// if 0 the iLCD is not restricted
	dword connected_ip;				// ip to which the iLCD is currently connected
      									// if 0 the iLCD is not connected
} tcp_info_2;

#ifdef LINUX
#pragma pack()
#else
#pragma option -a
#endif

// states of the current connection
typedef enum
{
   st_closed = 0,
   st_waiting_ack,
   st_connected,
} state_t;

// different password modes
typedef enum
{
	pwm_none = 0,
	pwm_plaintext,
	pwm_crc32,
	pwm_md5
} password_t;

state_t state = st_closed;
byte buffer[BUFF_SIZE] ;
static char password[PW_LEN + 1] = "";

char startup[3] = { ACK, '#', ACK };
char c[3] = { CIN, '$', '\0'};

static dword FWriteTimeout = DEFAULT_TX_TIMEOUT;
static dword FReadTimeout = DEFAULT_RX_TIMEOUT;

#ifndef LINUX
int Eth_startWinsock(void);
#endif

void Eth_Calculate_MD5(byte *hash, byte *challenge, char *password);
void Eth_Calculate_CRC32(byte *hash, byte *challenge, char *password);
//---------------------------------------------------------------------------
#ifndef LINUX
int Eth_startWinsock(void)
{
   WSADATA wsa;
   return WSAStartup(0x22,&wsa);      	 // use winsocks version 2.2
}
#endif
//---------------------------------------------------------------------------
// send data over ethernet and the function quotes
// the command line introducer (0xAA)
//---------------------------------------------------------------------------
void Eth_tcp_send(byte *data, int size)
{
 	int i;
   char intr = CIN;

   for (i = 0; i < size; i++)
   {
    	if (data[i] == (byte) CIN)
      {
         send(tcpsocket,&intr,1,0);
      }
      send(tcpsocket,&data[i],1,0);
   }
}
//---------------------------------------------------------------------------
// receive data with timeout
//---------------------------------------------------------------------------
int Eth_recvTimeout(SOCKET s, byte *buf, int len, int timeout)
{
   fd_set fds;
   int n;
   struct timeval tv;

   // set up the file descriptor set
   FD_ZERO(&fds);
   FD_SET(s, &fds);

   // set up the struct timeval for the timeout
   tv.tv_sec = timeout  / 1000;
   tv.tv_usec = (timeout % 1000) * 1000;

   // wait until timeout or data received
	n = select(s + 1, &fds, NULL, NULL, &tv);
	
	if (n == 0)
		return -2; // timeout!
		
	if (n == -1)
    	return -1; // error

   // data must be here, so do a normal recv()
   return recv(s, buf, len, 0);
}
//---------------------------------------------------------------------------
// Initialise sockets
//---------------------------------------------------------------------------
bool Eth_Init(void)
{
#ifndef LINUX
	if(Eth_startWinsock() != 0)
	{
		return Error_SetErr(ETH_ERR_WSA_LOAD);
	}
#endif

	udpsocket = socket(AF_INET, SOCK_DGRAM,0);  // initialise udp socket
	
	if (udpsocket == INVALID_SOCKET)
	{
		return Error_SetErr(ETH_ERR_CREATE_UDP);
	}

	tcpsocket = socket(AF_INET,SOCK_STREAM,0);  // initialise tcp socket
	
	if (tcpsocket == INVALID_SOCKET)
	{
		return Error_SetErr(ETH_ERR_CREATE_TCP);
	}
   return true;
}
//---------------------------------------------------------------------------
// returns true if an iLCD device is found. If there are more than one
// iLCD devices found an error will return
// The required device details are written to the tcp_info struct
//---------------------------------------------------------------------------
bool Eth_FindDevices(void)
{
	SOCKADDR_IN addr, echoAddr;
	fd_set fds;
	bool bOptVal = true;
	int bOptLen = sizeof(bool);
	int rc, n, count = 0;
	SOCKLEN echoAddrLen = sizeof(SOCKADDR_IN);
	tcp_info_2 tcp_info;
	struct timeval tv;

	// set the socket options for broadcast
	if (setsockopt(udpsocket, SOL_SOCKET, SO_BROADCAST, (char*) &bOptVal, bOptLen) != 0)
	{
		Error_SetErr(ETH_ERR_SET_SOCK_OPTION);
  		return false;
	}

	// broadcast address
	addr.sin_family=AF_INET;
	addr.sin_port=htons(PORT);
	addr.sin_addr.s_addr=inet_addr("255.255.255.255");

	rc = sendto(udpsocket, magicnumber, SIZE_MAGICNUMBER, 0, (SOCKADDR *) &addr, sizeof(SOCKADDR_IN));
	
	if (rc != SIZE_MAGICNUMBER)
	{
	  return Error_SetErr(ETH_ERR_SEND_UDP_BROADCAST);
	}

   FD_ZERO(&fds);
   FD_SET(udpsocket, &fds);

	tv.tv_sec = UDP_TIMEOUT  / 1000;
   tv.tv_usec = (UDP_TIMEOUT % 1000) * 1000;

	// waiting for device respons
	do
	{
		n = select(udpsocket + 1, &fds, NULL, NULL, &tv);
		
      if (n == 0) // no more device responed
      {
      	break;
      }
      if (n == -1) // error occured
      {
         return Error_SetErr(ETH_ERR_RECEIVE);
      }

		rc = recvfrom(udpsocket, buffer, BUFF_SIZE, 0, (SOCKADDR *) &echoAddr, &echoAddrLen);

		if (rc == (int) sizeof(tcp_info_1) || rc == (int) sizeof(tcp_info_2))
		{
			memcpy((char *) &tcp_info, buffer, rc);

			if (rc == (int) sizeof(tcp_info_1))
			{
				tcp_info.restricted_ip = 0;
				tcp_info.connected_ip = 0;
			}

			if (tcp_info.magic_no == SWAPPED_EEPROM_SIGNATURE)
			{
				count ++; // number of devices found
				sendAddr = echoAddr;	//device responded
				sendAddrLen = echoAddrLen;
			}
		}
	} while( rc > 0);

	if (count == 0)
	{
		return Error_SetErr(ETH_ERR_NO_DEVICES_FOUND);
	}
	else if (count == 1)
	{
		// check if device is already connected with a different application or PC
		
		if (tcp_info.connected_ip == 0)     
			return true;
		else
			return Error_SetErr(ETH_ALREADY_CONNECTED);
	}
	else if(count > 1)
	{
		return Error_SetErr(ETH_ERR_MORE_DEVICES_FOUND);
	}

	return Error_SetErr(ETH_ERR_UDP_RESPONSE);
}
//---------------------------------------------------------------------------
// open connection
//---------------------------------------------------------------------------
bool Eth_Open(void)
{
	long rc;
	byte hash[16];

	// looking for devices
	if (!Eth_FindDevices())
	{
		return false;
	}

	// start tcp connection
	rc = connect(tcpsocket,(SOCKADDR*)&sendAddr,sendAddrLen);

	if (rc == -1)
	{
		return Error_SetErr(ETH_ERR_TCP_CONNECT);
	}
	else if (rc == -2)
	{
		return Error_SetErr(ETH_ERR_RECEIVE_TIMEOUT);
	}

	do
	{
		// receive data
		if ((rc = Eth_recvTimeout(tcpsocket, buffer, BUFF_SIZE, FReadTimeout)) <= 0)
		{
			if (rc == 0)
			{
				Error_SetErr(ETH_ERR_NOT_CONNECTED);
			}
			else if (rc == -1)
			{
				Error_SetErr(ETH_ERR_TCP_CONNECT);
			}
			else if (rc == -2)
			{
				Error_SetErr(ETH_ERR_RECEIVE_TIMEOUT);
			}

			state = st_closed;
			return false;
		}

		if (state == st_connected)
		{
			return true;  // connection already open
		}
		else if (state == st_closed)
		{
			// no password required
			if (memcmp(buffer, startup + 1, 2) == 0)
			{
				state = st_connected;
				
				// connected
				return true;
			}

			// password required
			if (buffer[0] == '$')
			{
				if (strlen(Eth_getPassword()) == 0)
				{
					return Error_SetErr(ETH_ERR_NO_PW);  // no password spezified
				}
				else
				{
					c[2] =  buffer[1];

					send(tcpsocket,c,3,0);

					if (buffer[1] == pwm_plaintext)
					{
						// send the password in plaintext
						Eth_tcp_send((byte *) Eth_getPassword(), strlen(Eth_getPassword()) + 1);
					}
					else if (buffer[1] == pwm_crc32)
					{
						// calculate the crc32 and send it
						Eth_Calculate_CRC32(hash, buffer + 2, Eth_getPassword());
						Eth_tcp_send(hash, 4);
					}
					else if (buffer[1] == pwm_md5)
					{
						// calculate the md5 hash and send it
						Eth_Calculate_MD5(hash, buffer + 2, Eth_getPassword());
						Eth_tcp_send(hash, 16);
					}
					state = st_waiting_ack;
				}
			}
		}
		else if (state == st_waiting_ack)
		{
			if (memcmp(buffer, startup, 3) == 0)
			{
				// right password
				state = st_connected;
				return true;
			}
			else if (buffer[0] == NACK)
			{
				// wrong password
				Eth_Close();
				return Error_SetErr(ETH_ERR_WRONG_PW);
			}
	  }
	} while (rc > 0);

	return false;
}
//---------------------------------------------------------------------------
// close the connection
//---------------------------------------------------------------------------
bool Eth_Close(void)
{
	state = st_closed;
#ifdef LINUX
	close(tcpsocket);
#else
	closesocket(tcpsocket);
#endif
	return true;
}
//---------------------------------------------------------------------------
// set write and read timeout
//---------------------------------------------------------------------------
void Eth_SetTimeouts(dword rx_tout, dword tx_tout)
{
	FReadTimeout = max(rx_tout, 200);   // minimum read timeout is 200ms
	FWriteTimeout = tx_tout;
}
//---------------------------------------------------------------------------
// check if device is open or not
//---------------------------------------------------------------------------
bool Eth_IsDevOpen(void)
{
	if (state == st_connected)
		return true;
	else
		return false;
}
//---------------------------------------------------------------------------
// receive one character
//---------------------------------------------------------------------------
bool Eth_ReceiveChar(unsigned char *c)
{
	long rc;

	rc = Eth_recvTimeout(tcpsocket, c, 1, FReadTimeout);
	
	if (rc == 1)
	{
		return true;
	}
	else
	{
		if (rc == 0)
		{
			Error_SetErr(ETH_ERR_NOT_CONNECTED);
		}
		else if (rc == -1)
		{
			Error_SetErr(ETH_ERR_TCP_CONNECT);
		}
		else if (rc == -2)
		{
			Error_SetErr(ETH_ERR_RECEIVE_TIMEOUT);
		}

		state = st_closed;
		return false;
	}
}
//---------------------------------------------------------------------------
// transmit data
//---------------------------------------------------------------------------
bool Eth_Transmit(byte *data, dword length)
{
	 send(tcpsocket,data,length,0);
	 return true;
}
//---------------------------------------------------------------------------
// currently not implemented with timeout
//---------------------------------------------------------------------------
bool Eth_TransmitWTimeout(byte *data, dword length, word timeout)
{
	 return Eth_Transmit(data, length);
}
//---------------------------------------------------------------------------
// calculates the md5 hash of the challange + passowrd
//---------------------------------------------------------------------------
void Eth_Calculate_MD5(byte *hash, byte *challenge, char *password)
{
	MD5_CTX mdContext;

	MD5Init(&mdContext);
	MD5Update(&mdContext, challenge, SIZEOFCHALLENGE);
	MD5Update(&mdContext, (byte *) password, strlen(password));
	MD5Final(hash, &mdContext);
}
//---------------------------------------------------------------------------
// calculates the crc32 of the challange + passowrd
//---------------------------------------------------------------------------
void Eth_Calculate_CRC32(byte *hash, byte *challenge, char *password)
{
	unsigned int *crc = (unsigned int *) hash;

	*crc = crc32_init();
	*crc = crc32(*crc, challenge, SIZEOFCHALLENGE);
	*crc = crc32(*crc, (byte *) password, strlen(password));
	*crc = crc32_final(*crc);
}
//---------------------------------------------------------------------------
char *Eth_getPassword(void)
{
	return password;
}
//---------------------------------------------------------------------------
bool Eth_setPassword(char *pw)
{
	if (strlen(pw) > PW_LEN)
	{
		return false;
	}
	else
	{
      strcpy(password, pw);
	}

	return true;
}
//---------------------------------------------------------------------------
void Eth_End(void)
{
#ifndef LINUX
	WSACleanup();
#endif
}
//---------------------------------------------------------------------------



