//---------------------------------------------------------------------------
/*
 * usb.c
 *
 * Linux driver for Demmel iLCD DPC3080 USB interface, based on libusb 1.0
 *
 * Written by Arlet Ottens <arlet@c-scape.nl>
 */

//#define DEBUG

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <string.h>
#include <pty.h>
#include <utmp.h>
#include <poll.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <libusb-1.0/libusb.h>
#include "usb.h"

/*
 * The interface that we use. The iLCD only uses interface 1.
 */
#define HID_INTERFACE   0
#define BULK_INTERFACE  1

#define HID_REPORT_SET 0x09

/*
 * The vendor/product ID of the iLCD
 */
#define PRODUCT "iLCD Controller"
#define PRODUCT_LEN sizeof(PRODUCT)

/*
 * Size of the IRQ/CTRL buffer
 */
#define IRQ_SIZE 64
#define CTRL_SIZE 64
#define BULK_SIZE 64

static int use_hid;

static int tx_timeout = 100;
static int rx_timeout = 100;

static libusb_context *context;

/*
 * USB device list
 */

static libusb_device_handle *lcd;

bool USB_TransmitBulk(byte *data, DWORD length);

//---------------------------------------------------------------------------
/*
 * search the bus for Demmel iLCD device
 * return device handle if found, otherwise NULL.
 */
libusb_device_handle *search_device(void)
{
	libusb_device ** list = 0;
	ssize_t cnt = libusb_get_device_list(NULL, &list);
	ssize_t i = 0;
	unsigned char product[PRODUCT_LEN + 1];
	libusb_device_handle *handle;

	char buf[20];
	int saved_stderr = dup(2);

	for ( i = 0; i < cnt; i++ )
	{
		libusb_device *dev = list[i];
		struct libusb_device_descriptor dev_desc;
		struct libusb_config_descriptor *conf_desc;
		int res = libusb_get_device_descriptor(dev, &dev_desc);
		FILE *fp;

		if (res < 0)
		{
			Error_SetErr(USB_ERR_ENUM_DEV_SET);
			return 0;
		}

		// set stderr to null to avoid displaying permission failures

		fp = freopen("/dev/null", "w", stderr);

		res = libusb_open(dev, &handle);

		sprintf(buf, "/dev/fd/%d", saved_stderr);
		fp = freopen(buf, "w", stderr);

		// Permission etc. failed -> try next device
		if (res)
			continue;

		res = libusb_get_string_descriptor_ascii(handle, dev_desc.iProduct, product, sizeof(product));

		if (res > 0 && strcmp((char *) product, PRODUCT) == 0)
		{

			// found suitable device, use HID mode if it only has 1 interface 
			res = libusb_get_config_descriptor(dev, 0, &conf_desc);
			use_hid = (conf_desc->bNumInterfaces == 1); 
			libusb_free_config_descriptor( conf_desc );
			libusb_free_device_list(list, 1);
			return handle;
		}
		libusb_close(handle);
	}
	libusb_free_device_list(list, 1);
	return 0;
}

//---------------------------------------------------------------------------
/*
 * USB_Open: open the USB device. Return true if successful
 */
bool USB_Open(void)
{
	int res;
	int interface;

	if (libusb_init(&context))
	{
		Error_SetErr(USB_ERR_LIBUSB_INIT);
		return false;
	}

	libusb_set_debug(context, 0);
	lcd = search_device();

	if (!lcd)
	{
		Error_SetErr(USB_ERR_NO_DEVICES_FOUND);
		return false;
	}

	interface = use_hid ? HID_INTERFACE : BULK_INTERFACE;

	if (libusb_kernel_driver_active(lcd, interface))
		libusb_detach_kernel_driver(lcd, interface);

	res = libusb_claim_interface(lcd, interface);

	if( !use_hid )
	    USB_TransmitBulk( 0, 0 ); // transmit empty data to signal bulk mode
	return res >= 0;
}

//---------------------------------------------------------------------------
/*
 * USB_TransmitHid: Transmit a block of data through HID interface. The data is broken up into
 * control packets, with the first byte indicating the length
 */
bool USB_TransmitHid(byte *data, DWORD length)
{
	int burst;
	int err;
	unsigned char buf[CTRL_SIZE];

	while (length > 0)
	{
		burst = min(length, CTRL_SIZE - 1);
		buf[0] = burst;
		memcpy(buf + 1, data, burst);
		err =
			libusb_control_transfer(lcd, LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE,
					HID_REPORT_SET, 0x200, 0, buf, burst + 1, tx_timeout);

		if (err < 0)
		{
			Error_SetErr(USB_ERR_WRITE);
			return false;
		}

		data += burst;
		length -= burst;
	}
	return true;
}

//---------------------------------------------------------------------------
/*
 * USB_TransmitBulk: Transmit a block of data through bulk interface.
 */
bool USB_TransmitBulk(byte *data, DWORD length)
{
	int burst;
	int err;

#ifdef DEBUG
	fprintf( stderr, "transmit bulk len %d\n", (int) length );
	int i;
	for( i = 0; i < length; i++ )
	    fprintf( stderr, "%02x%c", data[i], i % 16 == 15 ? '\n' : ' ' );
	fprintf( stderr, "\n" );
#endif
	while( length > 0 )
	{
		err = libusb_bulk_transfer(lcd, LIBUSB_ENDPOINT_OUT | 2, data, length, &burst, tx_timeout);

		if (err < 0)
		{
			Error_SetErr(USB_ERR_WRITE);
			return false;
		}
		data += burst;
		length -= burst;
	}
	return true;
}

bool USB_Transmit(byte *data, DWORD length)
{
	if( use_hid )
		return USB_TransmitHid( data, length );
	else
		return USB_TransmitBulk( data, length );
}

//---------------------------------------------------------------------------

/*
 * USB_ReceiveCharHid: read a single character through HID interface. 
 * If the buffer is empty, read a new interrupt packet from the device
 */
bool USB_ReceiveCharHid(unsigned char *c)
{
	static unsigned char rx_buf[IRQ_SIZE];
	static int rx_length = 0;
	static int rx_ptr = 0;
	int length;

	if (rx_ptr == rx_length)
	{
		libusb_interrupt_transfer(lcd, 1 | LIBUSB_ENDPOINT_IN, rx_buf, IRQ_SIZE, &length, rx_timeout);

		if (length == 0)
		{
			Error_SetErr(USB_ERR_RX_NO_DATA);
			return false;
		}
		rx_length = 1 + rx_buf[0];
		rx_ptr = 1;
	}

	if (rx_ptr == rx_length)
	{
		Error_SetErr(USB_ERR_RX_NO_DATA);
		return false;
	}

	*c = rx_buf[rx_ptr++];

	return true;
}

/*
 * USB_ReceiveCharBulk: read a single character through bulk interface. 
 * If the buffer is empty, read a new bulk packet from the device
 */
bool USB_ReceiveCharBulk(unsigned char *c)
{
	static unsigned char rx_buf[BULK_SIZE];
	static int rx_length = 0;
	static int rx_ptr = 0;
	int length;

	if (rx_ptr == rx_length)
	{
		libusb_bulk_transfer(lcd, 2 | LIBUSB_ENDPOINT_IN, rx_buf, sizeof(rx_buf), &length, rx_timeout);
#ifdef DEBUG
		fprintf( stderr, "got %d bytes\n", length );
#endif
		if (length == 0)
		{
			Error_SetErr(USB_ERR_RX_NO_DATA);
			return false;
		}
#ifdef DEBUG
		int i;
		for( i = 0; i < length; i++ )
			fprintf( stderr, "%02x%c", rx_buf[i], i % 16 == 15 ? '\n' : ' ' );
		fprintf( stderr, "\n" );
#endif
		rx_length = length;
		rx_ptr = 0;
	}

	if (rx_ptr == rx_length)
	{
		Error_SetErr(USB_ERR_RX_NO_DATA);
		return false;
	}

	*c = rx_buf[rx_ptr++];

	return true;
}

bool USB_ReceiveChar(unsigned char *c)
{
	if( use_hid )
		return USB_ReceiveCharHid( c );
	else
		return USB_ReceiveCharBulk( c );
}

//---------------------------------------------------------------------------
/*
 * USB_SetTimeouts:
 * set the Receive and Transmit timeouts in milliseconds
 */
void USB_SetTimeouts(dword rx_tout, dword tx_tout)
{
	rx_timeout = rx_tout;
	tx_timeout = tx_tout;
}

//---------------------------------------------------------------------------
/*
 * USB_Close: close the device, and free resources
 */
bool USB_Close(void)
{
	libusb_close(lcd);
	libusb_exit(context);
	return true;
}
//---------------------------------------------------------------------------
